/*
 * Copyright (C) 2012  Arnaud Bonatti  <arnaud.bonatti@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

const Clutter		= imports.gi.Clutter
const Gio			= imports.gi.Gio
const Gtk			= imports.gi.Gtk
const GLib			= imports.gi.GLib
const IconGrid		= imports.ui.iconGrid
const Lang			= imports.lang
const Main			= imports.ui.main
const Meta			= imports.gi.Meta
const Params		= imports.misc.params
const RunDialog		= imports.ui.runDialog
const Search		= imports.ui.search
const Shell			= imports.gi.Shell
const St			= imports.gi.St
const Util			= imports.misc.util
const ViewSelector	= imports.ui.viewSelector
const MAGIC_FISH_KEY= imports.ui.wanda.MAGIC_FISH_KEY

const MainFile		= imports.misc.extensionUtils.getCurrentExtension().imports.extension	// local file
const gettextDomain	= 'terminal-provider'
const _				= imports.gettext.domain(gettextDomain).gettext
const S_			= imports.gettext.domain('gnome-shell').gettext

// schemas
const dconfSchema	= 'org.gnome.shell.extensions.terminal-provider'

const TERM_SCHEMA	= 'org.gnome.desktop.default-applications.terminal'
const TERMNAME_KEY	= 'exec'		// no loading of 'exec-arg': use of "-x" for gnome-terminal and "-e" for others

const LOCK_SCHEMA	= 'org.gnome.desktop.lockdown'
const NO_SHELL_KEY	= 'disable-command-line'

// strings
const EXEC_TEXT		= "Launch command"
const TERM_TEXT		= "Exec in terminal"
const FILE_TEXT		= "Try opening file"
const FOLDER_TEXT	= "Open folder"
const LG_TEXT		= "Looking Glass"; let LG_TEXT_lc, LG_TEXT_lc_i18n
const LG_ABREV		= "lg"
const OVERVIEWTITLE	= "TERMINAL"				// Title in the Overview

// icons
const EXEC_ICON		= "system-run"
const TERM_ICON		= "gnome-terminal"
const FOLDER_ICON	= "folder"
const FILE_ICON		= "document-open" // or "text-x-generic"? urgh.
const DEVS_ICON		= "applications-development"

let TERMS

/* Remove <Alt>F2, one time */

const SHELLENTRYKEY = 'org.gnome.desktop.wm.keybindings'

let lockFilePath = null
let connectionAltF2 = null
let seSettings = null

function disableShellEntry() {
	seSettings = new Gio.Settings({ 'schema': SHELLENTRYKEY })
	MainFile.getSettings(dconfSchema).set_strv('shell-entry-keybinding', seSettings.get_strv('panel-run-dialog'))
	seSettings.set_strv('panel-run-dialog', [])		// for the control-center not showing the old keybinding

	// not possible to play directly with built-in keybindings (22-08-2012, -> 3.6), so lets play a little
	Meta.keybindings_set_custom_handler('panel-run-dialog', function() {})
	connectionAltF2 = seSettings.connect('changed::panel-run-dialog', function() { if (seSettings.get_strv('panel-run-dialog')!="") rEnableShellEntry(true) })	// TODO test "changed::panel-run-dialog"
}

function rEnableShellEntry(createLockFile) {
	Meta.keybindings_set_custom_handler('panel-run-dialog', function() { Main.getRunDialog().open() })
	seSettings.disconnect(connectionAltF2)
	
	if (createLockFile) {
		let file = Gio.file_new_for_path(lockFilePath)
		let out = file.replace(null, false, Gio.FileCreateFlags.NONE, null)
		Shell.write_string_to_stream (out, "true")
	} else
		seSettings.set_strv('panel-run-dialog', MainFile.getSettings(dconfSchema).get_strv('shell-entry-keybinding'))
	
	connectionAltF2 = null
	seSettings = null
}

/* Terminal Provider */
function TerminalExec() { this._init() }

TerminalExec.prototype = {
	__proto__: Search.SearchProvider.prototype,

	_init: function() {
		Search.SearchProvider.prototype._init.call(this, _(OVERVIEWTITLE))
		
		this._connectionTools = global.settings.connect('changed::development-tools', Lang.bind(this, function () {		// as in js/ui/runDialog.js:174
			this._enableInternalCommands = global.settings.get_boolean('development-tools')
		}))
		this._enableInternalCommands = global.settings.get_boolean('development-tools')
		
		this._appSystem = Shell.AppSystem.get_default()
		
		// TODO need a SIMPLE way to middle-click, not as in appDisplay.js:521; instead of using createResultActor (see searchDisplay.js:34), hack
	},

	getResultMetas: function(ids, callback) {
		let textureCache = St.TextureCache.get_default()
		
		let cb = new Array()
		switch (ids[0]) {
			case "open":
				this.file = Gio.file_new_for_path(this.path)		// used in this._openFile() even for dirs
				if (GLib.file_test(this.path, GLib.FileTest.IS_DIR))
					cb.push({ 'id': "open",	'name': _(FOLDER_TEXT),
						'createIcon': function(size) { return textureCache.load_icon_name(null, FOLDER_ICON, size)} })
				else {
					let icon = new Gio.FileIcon({ file: this.file })
					let thumbnail = this.file.query_info(Gio.FILE_ATTRIBUTE_THUMBNAIL_PATH, Gio.FileQueryInfoFlags.NONE, null).get_attribute_as_string(Gio.FILE_ATTRIBUTE_THUMBNAIL_PATH)		// ..._THUMBNAILING_FAILED fails?
					if (thumbnail)	cb.push({ 'id': "open",	'name': _(FILE_TEXT),
						'createIcon': function(size) { return textureCache.load_gicon(null, icon, size)} })
					else 			cb.push({ 'id': "open",	'name': _(FILE_TEXT),
						'createIcon': function(size) { return textureCache.load_icon_name(null, FILE_ICON, size)} })
				}
				
				cb.push({ 'id': "term",	'name': _(TERM_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, TERM_ICON, size)} },
						{ 'id': "exec",	'name': _(EXEC_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, EXEC_ICON, size)} })
				break
			case "lg":
				cb.push({ 'id': "lg",	'name': _(LG_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, DEVS_ICON, size)} })
				if (ids.length > 1) cb.push(
						{ 'id': "term",	'name': _(TERM_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, TERM_ICON, size)} },
						{ 'id': "exec",	'name': _(EXEC_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, EXEC_ICON, size)} })
				break
			case "exec":	// not all exec
				cb.push({ 'id': "exec",	'name': _(EXEC_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, EXEC_ICON, size)} },
						{ 'id': "term",	'name': _(TERM_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, TERM_ICON, size)} })
				break
			default:		// an exec sometimes gives the app name
				cb.push({ 'id': "exec",	'name': _(EXEC_TEXT),
					'createIcon': function(size) { return Shell.AppSystem.get_default().lookup_app(ids[0] + ".desktop").create_icon_texture(size) }},
						{ 'id': "term",	'name': _(TERM_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, TERM_ICON, size)} })
		}
		callback(cb)
	},
	
	getInitialResultSet: function(terms) {							// same as getSubsearchResultSet(...)
		this.terms = terms		// not .join(' ') here, 'words' are useful
		
		let text = terms.join(' ')
		this.path = text.charAt(0) == '/' ? text : GLib.build_filenamev([GLib.get_home_dir(), text.slice(0,2) == '~/' ? text.slice(2) : terms[0] == '~' ? text.slice(1) : text])
		text = text.toLowerCase()
		
		if (GLib.file_test(this.path, GLib.FileTest.EXISTS))
			this.searchSystem.pushResults(this, ["open", "term", "exec"])
		else if (this._enableInternalCommands
			&&( text == LG_TEXT_lc_i18n.substr(0, text.length) || text == LG_TEXT_lc.substr(0, text.length) || text == LG_ABREV.substr(0, text.length) ))
			{ if (terms[1]) this.searchSystem.pushResults(this, ["lg"])
			else this.searchSystem.pushResults(this, ["lg", "term", "exec"]) }
		else if (this._appSystem.lookup_app(terms[0] + ".desktop"))
			this.searchSystem.pushResults(this, [terms[0], "term"])
		else if (text == MAGIC_FISH_KEY)		// assume the MFK is lowercase
			this.searchSystem.pushResults(this, [])
		else
			this.searchSystem.pushResults(this, ["exec", "term"])
	},
	getSubsearchResultSet: function(previousResults, terms) {		// same as getInitialResultSet(terms)
		this.terms = terms		// not .join(' ') here, 'words' are useful
		
		let text = terms.join(' ')
		this.path = text.charAt(0) == '/' ? text : GLib.build_filenamev([GLib.get_home_dir(), text.slice(0,2) == '~/' ? text.slice(2) : terms[0] == '~' ? text.slice(1) : text])
		text = text.toLowerCase()
		
		if (GLib.file_test(this.path, GLib.FileTest.EXISTS))
			this.searchSystem.pushResults(this, ["open", "term", "exec"])
		else if (this._enableInternalCommands
			&&( text == LG_TEXT_lc_i18n.substr(0, text.length) || text == LG_TEXT_lc.substr(0, text.length) || text == LG_ABREV.substr(0, text.length) ))
			{ if (terms[1]) this.searchSystem.pushResults(this, ["lg"])
			else this.searchSystem.pushResults(this, ["lg", "term", "exec"]) }
		else if (this._appSystem.lookup_app(terms[0] + ".desktop"))
			this.searchSystem.pushResults(this, [terms[0], "term"])
		else if (text == MAGIC_FISH_KEY)		// assume the MFK is lowercase
			this.searchSystem.pushResults(this, [])
		else
			this.searchSystem.pushResults(this, ["exec", "term"])
	},
	activateResult: function(id) {
		if		(id == "exec")	this._execLine()
		else if (id == "term")	this._execTerm()
		else if	(id == "open")	this._openFile()
		else /* (id == "lg") */	Main.createLookingGlass().toggle()
	},
	dragActivateResult: function(id, params) {		// same as activateResult but for d’n’d, see searchDisplay.js:95
		params = Params.parse(params, { workspace: -1, timestamp: 0 })		// call on workspaceThumbnail.js:477
		let launchWorkspace = global.screen.get_workspace_by_index(params.workspace)
		launchWorkspace.activate(global.get_current_time())
		
		if		(id == "exec")	this._execLine()
		else if (id == "term")	this._execTerm()
		else if	(id == "open")	this._openFile()
		else /* (id == "lg") */	Main.createLookingGlass().toggle()
		
		Main.overview.hide()		// Overview auto-hide with activateResult() but not with dragActivateResult(), even for Looking Glass...
	},
	_parseCommandLine: function(recursive) {
		let i, j
		let words = new Array()
		let words_are_path = 0		// = words.length
		
		for (i=0; i<this.terms.length && recursive; i++) {
			if (this.terms[i].slice(0,2) == '~/')	this.terms[i] = GLib.build_filenamev([GLib.get_home_dir(), this.terms[i].slice(2)])
			else if (this.terms[i] == '~') 			this.terms[i] = GLib.get_home_dir()
			
			let text
			for (j=this.terms.length; j>i; j--) {
				text = this.terms.slice(i,j).join(' ')
				if (GLib.file_test(text, GLib.FileTest.EXISTS)) {		// Glib.file_test("~/blabla", etc.) fails (they have been removed), Glib.file_test("blabla", etc.) is ok
					if (text[0] != '/') text = GLib.build_filenamev([GLib.get_home_dir(), text])
					i = j - 1
					words_are_path += 1
					break	// the j loop
				}
			}
			words.push(text)
		}
		return [words, words_are_path]
	},
	_execLine: function() {
		let [words, number_of_paths] = this._parseCommandLine(true)
		
		if ( ( GLib.file_test(words[0], GLib.FileTest.IS_REGULAR) && !GLib.file_test(words[0], GLib.FileTest.IS_EXECUTABLE))
			|| GLib.file_test(words[0], GLib.FileTest.IS_DIR)) {
			if (number_of_paths == words.length) {
				let i = 0
				for (; i<number_of_paths; i++) this._openFile(Gio.file_new_for_path(words[i]))
				global.log(words[0] + " is not executable, opening all the files.")
			} else
				global.log("Error when trying to understand the command. Do you know that\n              " + words[0] + " is not executable?")
		} else
			Util.spawn(words)
	},
	_execTerm: function() {
		let i = 0, error
		do {
			error = null
			try {
				if (TERMS[i+1])	Util.trySpawn([TERMS[i], (TERMS[i]=="gnome-terminal")?"-x":"-e"].concat(this.terms))
				else			Util.spawn([TERMS[i], (TERMS[i]=="gnome-terminal")?"-x":"-e"].concat(this.terms))
			} catch (err) {
				error = err
				global.log(TERMS[i] + " not found (extension306.js:273). Trying " + TERMS[i+1] + ".")
				i += 1
			}
		} while (error)
	},
	_openFile: function(file) {
		if (!file) file = this.file		// see line 127
		try { Gio.app_info_launch_default_for_uri(file.get_uri(), global.create_app_launch_context()) }
		catch (e) {
			// The exception from gjs contains an error string like:
			//		Error invoking Gio.app_info_launch_default_for_uri: No application
			//		is registered as handling this file
			// We are only interested in the part after the first colon.
			let message = e.message.replace(/[^:]*: *(.+)/, '$1');
			global.log(message)
		}
	},
	byebye: function() {
		global.settings.disconnect(this._connectionTools)
		this._connectionTools = 0
	}
}

/* Autocompletion */

function Autocompletor(params) { this._init(params) }		// has an St.Entry with autocompletion as actor

Autocompletor.prototype = {
	_init: function(params) {
		// params = Params.parse(params, { hint_text: _("Type command..."), track_hover: false, can_focus: true })
		this._entry = new St.Entry(params)
		this._pathCompleter = new Gio.FilenameCompleter()
		this._commandCompleter = new RunDialog.CommandCompleter()
		this._commCompUpdateConnection = Main.overview.connect('shown', Lang.bind(this._commandCompleter, this._commandCompleter.update))

		this._haveToPreload = true
		this._textInsertConnection = this._entry.clutter_text.connect('insert-text', Lang.bind(this, function(clutter_text, chars, is_empty) {
			if (chars.indexOf('/')!=-1) this._haveToPreload = true
		}))
		this._textDeleteConnection = this._entry.clutter_text.connect('delete-text', Lang.bind(this, function() {
			this._haveToPreload = true
		}))
		this._cursorEventConnection = this._entry.clutter_text.connect('cursor-event', Lang.bind(this, function(clutter_text, geometry) {
			let _cursorPosition = clutter_text.get_cursor_position()
			// -1 is before the last character for slice, and after for insert_text(), cursor_get_position() and others, both used for completion...
			this._cursorPosition = (_cursorPosition == -1) ? clutter_text.get_text().length : _cursorPosition
		
			if (!this._haveToPreload) return
			let text = clutter_text.get_text().slice(0, this._cursorPosition)
			if (text.indexOf('/') != -1) this._pathCompletion(text.concat('a'))		// GFilenameCompleter (this._pathCompleter) loads content of parent directory; so add a fake thing ('a') to permit preloading when last character is '/'
			else this._pathCompleter.get_completion_suffix('~/a')
			this._haveToPreload = false
		}))

		this._autoCompletionConnection = this._entry.clutter_text.connect('key-press-event', Lang.bind(this, function(clutter_text, event) {	// inspired by js/ui/runDialog.js:243
			if(event.get_key_symbol() == Clutter.Tab) {
				let postfix = this._getCompletion()
				if (postfix != null && postfix.length > 0) {
					clutter_text.insert_text(postfix, this._cursorPosition)
					clutter_text.set_cursor_position(this._cursorPosition + postfix.length)
				}
				return true
			}
			return false
		}))
	},
	
	byebye: function() {
		this._entry.clutter_text.disconnect(this._textInsertConnection)
		this._entry.clutter_text.disconnect(this._textDeleteConnection)
		this._entry.clutter_text.disconnect(this._cursorEventConnection)
		
		Main.overview.disconnect(this._commCompUpdateConnection)
		this._commandCompleter = null
		this._pathCompleter = null
		
		this._entry.destroy()
	},
	
	_pathCompletion: function(text) {
		let index
		do {
			let test = text.slice(0,2) == '~/' ? 2 : 0
			if (GLib.file_test(text.slice(test, text.lastIndexOf('/') + 1), GLib.FileTest.IS_DIR)) {		// Glib.file_test("~/blabla", etc.) fails, Glib.file_test("blabla", etc.) is ok ; pathCompleter.get_completion_suffix("~/blabla") is ok
				if (text[0] == '/')	return this._pathCompleter.get_completion_suffix(text)
				if (!test) text = '~/'.concat(text)
				return this._pathCompleter.get_completion_suffix(text)
			}
			index = text.indexOf(' ') + 1
			text = text.slice(index)
		} while (index && (text.indexOf('/') != -1))	// case [abc/def without ~/abc/def] and case [abc def without ~/abc\ def]
		return false									// case [~/def] is tested by _getCompletion, same code as for when there isn't '/' initial text
	},
	
	_getCompletion: function() {
		let text = this._entry.clutter_text.get_text().slice(0, this._cursorPosition)
		let index = text.lastIndexOf('/') + 1
		if (index) {
			let test = this._pathCompletion(text)
			if (test) return test
			text = text.slice(index)
		}
		do {
			index = text.lastIndexOf(' ') + 1
			text = text.slice(index)
			let test = this._commandCompleter.getCompletion(text)		// TODO: fix the "looking glade" autocompletion problem
			if (test) return test
		} while (index)
		return false
	}
}

/* Settings Guard */
let settingsGuard = null

function SettingsGuard() { this._init() }

SettingsGuard.prototype = {
	_init: function() {
		this.TermProvider = null
		
		this._termSettings = new Gio.Settings({ schema: TERM_SCHEMA })
		this._connectionTerm = this._termSettings.connect('changed::' + TERMNAME_KEY, this._listTerminals)	// see line 39
		this._listTerminals(this._termSettings)
		
		this._lockSettings = new Gio.Settings({ schema: LOCK_SCHEMA })
		this._connectionLock = this._lockSettings.connect('changed::' + NO_SHELL_KEY, Lang.bind(this, this._syncTermProvider))
		this._syncTermProvider()
	},
	
	// Terminals
	_listTerminals: function(_termSettings) {
		TERMS = new Array(_termSettings.get_string(TERMNAME_KEY))
		if (TERMS.indexOf("gnome-terminal") == -1)		TERMS.push("gnome-terminal")
		if (TERMS.indexOf(GLib.getenv('TERM')) == -1)	TERMS.push(GLib.getenv('TERM'))
		if (TERMS.indexOf("xfce4-terminal") == -1)		TERMS.push("xfce4-terminal")
		if (TERMS.indexOf("xterm") == -1)				TERMS.push("xterm")
	},
	
	// Autocompletion
	enableAutocompletion: function() {
		let overview = Main.overview
		let group = overview._group
		let dash = overview._dash
		
		// Search Entry
		this._autocompletor = new Autocompletor({ name: 'searchEntry', hint_text: S_("Type to search..."), track_hover: true, can_focus: true })
		this._newSearchEntry = this._autocompletor._entry
		this._newSearchEntry.set_width(0.44 * Main.layoutManager.primaryMonitor.width)
		
		this._oldSearchEntry = overview._searchEntry
		
		group.remove_actor(overview._searchEntry)
		overview._searchEntry = this._newSearchEntry
		group.add_actor(overview._searchEntry)
		
		// View Selector
		this._newViewSelector = new ViewSelector.ViewSelector(overview._searchEntry, dash.showAppsButton)
		this._oldViewSelector = overview._viewSelector
		
		dash.actor.remove_constraint(overview._viewSelector.constrainHeight)
		group.remove_actor(dash.actor)
		group.remove_actor(overview._viewSelector.actor)
		overview._viewSelector = this._newViewSelector
		group.add_actor(overview._viewSelector.actor)
		group.add_actor(dash.actor)
		dash.actor.add_constraint(overview._viewSelector.constrainHeight)
		
		overview._relayout()
	},
	disableAutocompletion: function() {
		let overview = Main.overview
		let group = overview._group
		let dash = overview._dash
		
		// Search Entry
		overview._group.remove_actor(overview._searchEntry)
		overview._searchEntry = this._oldSearchEntry
		overview._group.add_actor(overview._searchEntry)
		
		this._autocompletor.byebye()
		this._autocompletor = null
		
		// View Selector
		dash.actor.remove_constraint(overview._viewSelector.constrainHeight)
		group.remove_actor(dash.actor)
		group.remove_actor(overview._viewSelector.actor)
		overview._viewSelector = this._oldViewSelector
		group.add_actor(overview._viewSelector.actor)
		dash.actor.add_constraint(overview._viewSelector.constrainHeight)
		
		this._newViewSelector.actor.destroy()
		this._newViewSelector = null
		group.add_actor(dash.actor)
		
		overview._relayout()
	},
	
	// TermProvider
	_syncTermProvider: function() {
		if (!this._lockSettings.get_boolean(NO_SHELL_KEY)) {
			if (!this.TermProvider) {
				this.enableAutocompletion()
		
				this.TermProvider = new TerminalExec()
				Main.overview.addSearchProvider(this.TermProvider)
			}
		}
		else this._disableTermProvider()
	},
	_disableTermProvider: function() {
		if (this.TermProvider) {
			this.disableAutocompletion()
		
			Main.overview.removeSearchProvider(this.TermProvider)
			this.TermProvider.byebye()
			this.TermProvider = null
		}
	},
	
	// Disable
	byebye: function() {
		this._disableTermProvider()
		
		this._lockSettings.disconnect(this._connectionLock); this._connectionLock = null; this._lockSettings = null
		this._termSettings.disconnect(this._connectionTerm); this._connectionTerm = null; this._termSettings = null
	}
}

/* Extensions API */
function init(metadata) {
	MainFile.initTranslations(gettextDomain)
	lockFilePath = GLib.build_filenamev([metadata.path, "managed"])
	
	LG_TEXT_lc		= LG_TEXT.toLowerCase()
	LG_TEXT_lc_i18n	= _(LG_TEXT).toLowerCase()
}
function enable() {
	settingsGuard = new SettingsGuard()
	
	// The extension disables the <Alt>F2 keybinding until the user manually re-set it. If he does, a file is created in the extension's folder that "lock" the behaviour until the extension is removed. The file should be deleted at each manual unactivation of the extension, but that looks like impossible with the actual API.
	if (!GLib.file_test(lockFilePath, GLib.FileTest.EXISTS)) disableShellEntry()
}
function disable() {
	if (!GLib.file_test(lockFilePath, GLib.FileTest.EXISTS)) rEnableShellEntry()
	
	settingsGuard.byebye()
	settingsGuard = null
}
